import random
import time
from itertools import product

import pytest

from atqo import parallel_map
from atqo.distributed_apis import DIST_API_MAP


def add2(x):
    return x + 2


def extstr(x):
    return f"{x}-add"


def div(x):
    return 10 / x


@pytest.mark.parametrize(
    ["fun", "inl", "dapi"],
    product(
        [add2, extstr, div],
        [[1, 2, 3, 4, 5], ["a", 10, "b", 3], [None, 2, 0]],
        DIST_API_MAP.keys(),
    ),
)
def test_batch(fun, inl, dapi):
    res = []
    exs = []
    for x in inl:
        try:
            res.append(fun(x))
        except Exception as e:
            exs.append(e)

    map_outs = [
        list(parallel_map(fun, inl, dapi, raise_errors=False, verbose=True)),
        parallel_map(fun, iter(inl), dapi, raise_errors=False),
    ]

    for mout in map_outs:
        mres = []
        mex = []
        for o in mout:
            if isinstance(o, Exception):
                mex.append(o)
            else:
                mres.append(o)
        assert sorted(mres) == sorted(res)
        for _w in [str, type]:
            assert set(map(_w, mex)) == set(map(_w, exs))


def test_extras():

    with pytest.raises(TypeError):
        list(parallel_map(add2, [1, 2, 3, "X", "Y"], batch_size=3, pbar=True))


def slow_fun(x):  # pragma: no cover
    time.sleep(random.random() / 100)
    return x


def test_simple():
    out = sorted(parallel_map(slow_fun, range(100), pbar=True, verbose=True))
    assert out == list(range(100))

    def g():
        for i in range(10):
            yield i

    assert sorted(parallel_map(slow_fun, g(), pbar=True)) == list(range(10))



import time

import pytest

from atqo.bases import ActorBase
from atqo.core import Scheduler, SchedulerTask
from atqo.distributed_apis import DEFAULT_MULTI_API
from atqo.resource_handling import Capability, CapabilitySet


class Actor(ActorBase):
    def __init__(self) -> None:  # pragma: no cover
        self.e = 1

    def consume(self, task_arg):  # pragma: no cover
        return self.e + task_arg


class ActorSlow(Actor):
    def __init__(self) -> None:  # pragma: no cover
        time.sleep(0.5)
        self.e = 20


cap1 = Capability({"A": 1})
cap2 = Capability({"B": 1})


@pytest.fixture
def test_scheduler():
    return Scheduler(
        actor_dict={
            CapabilitySet([cap1]): Actor,
            CapabilitySet([cap2]): ActorSlow,
        },
        resource_limits={"A": 3, "B": 2},
        distributed_system=DEFAULT_MULTI_API,
        verbose=True,
    )


def test_runoff(test_scheduler: Scheduler):
    def _get_st(cap, n):
        return [SchedulerTask(i, [cap]) for i in range(n)]

    class _BP:
        def __init__(self) -> None:
            self.i = -1
            self._batches = [
                _get_st(cap1, 10),
                _get_st(cap2, 3),
                _get_st(cap1, 5),
                _get_st(cap2, 10),
            ]

        def __call__(self):
            self.i += 1
            if self.i >= len(self._batches):
                return []
            return self._batches[self.i]

    out = list(test_scheduler.process(_BP(), min_queue_size=5))
    test_scheduler.join()
    assert len(out) == 28


def test_slow_start(test_scheduler: Scheduler):

    rtasks = [SchedulerTask(i, [cap1]) for i in range(10)]
    stasks = [SchedulerTask(i, [cap2]) for i in range(2)]

    test_scheduler.refill_task_queue(rtasks + stasks)
    assert sorted(test_scheduler.join()) == [
        *range(1, 11),
        *range(20, 22),
    ]



from functools import partial
from itertools import islice
from multiprocessing import cpu_count

from .bases import ActorBase
from .core import Scheduler, SchedulerTask
from .distributed_apis import DEFAULT_MULTI_API
from .resource_handling import Capability, CapabilitySet

_RES = "CPU"
_CAP = Capability({_RES: 1})
_Task = partial(SchedulerTask, requirements=[_CAP])


class BatchProd:
    def __init__(self, iterable, batch_size, mapper=_Task) -> None:
        self._size = batch_size
        self._it = iter(iterable)
        self._mapper = mapper

    def __call__(self):
        return [*map(self._mapper, islice(self._it, self._size))]


class ActWrap(ActorBase):
    def __init__(self, fun) -> None:
        self._f = fun

    def consume(self, task_arg):
        return self._f(task_arg)


def get_simp_scheduler(n, fun, dist_sys, verbose, max_runs) -> Scheduler:
    # TODO this will not work
    ActWrap.restart_after = max_runs

    return Scheduler(
        actor_dict={CapabilitySet([_CAP]): partial(ActWrap, fun=fun)},
        resource_limits={_RES: n},
        distributed_system=dist_sys,
        verbose=verbose,
    )


def parallel_map(
    fun,
    iterable,
    dist_api=DEFAULT_MULTI_API,
    batch_size=None,
    min_queue_size=None,
    workers=None,
    raise_errors=True,
    verbose=False,
    pbar=False,
    restart_after=float("inf"),
):
    nw = workers or cpu_count()
    batch_size = batch_size or nw * 5
    min_queue_size = min_queue_size or batch_size // 2

    pinger = get_pinger(iterable) if pbar else lambda: None
    scheduler = get_simp_scheduler(nw, fun, dist_api, verbose, restart_after)

    out_iter = scheduler.process(
        batch_producer=BatchProd(iterable, batch_size),
        min_queue_size=min_queue_size,
    )
    try:
        for e in out_iter:
            if raise_errors and isinstance(e, Exception):
                raise e
            pinger()
            yield e
    finally:
        scheduler.join()


def get_pinger(iterable):
    from tqdm import tqdm

    try:
        total = len(iterable)
    except Exception:
        total = None

    return tqdm(total=total).update



from functools import partial
from operator import add, sub


def dic_merge(d1: dict, d2: dict, _def, fun):
    return {k: fun(d1.get(k, _def), d2.get(k, _def)) for k in [*d1, *d2]}


def dic_val_filt(dic):
    return {k: v for k, v in dic.items() if v}


sumdict = partial(dic_merge, _def=0, fun=add)
subdict = partial(dic_merge, _def=0, fun=sub)



import pytest

from atqo import ActorBase, Capability, CapabilitySet, Scheduler, SchedulerTask
from atqo.distributed_apis import DEFAULT_DIST_API_KEY
from atqo.simplified_functions import BatchProd

LIMIT_DIC = {"A": 3}

cap1 = Capability({"A": 1})
cap2 = Capability({"A": 1})


class Actor(ActorBase):
    def consume(self, task_arg):
        return f"done-{task_arg}"


@pytest.fixture
def test_scheduler():
    actor_dict = {
        CapabilitySet([cap2, cap1]): Actor,
    }

    return Scheduler(
        actor_dict=actor_dict,
        resource_limits=LIMIT_DIC,
        distributed_system=DEFAULT_DIST_API_KEY,
        verbose=True,
    )


class ListProd:
    def __init__(self, list_base, batch_size) -> None:
        self._size = batch_size
        self._list = list_base
        self._i = 0

    def __call__(self):
        out = []
        for _ in range(self._size):
            try:
                out.append(self._list[self._i])
                self._i += 1
            except IndexError:
                break
        return out


def test_over_actors(test_scheduler: Scheduler):

    tasks = [
        SchedulerTask("task1", requirements=[cap1]),
    ]

    out = list(test_scheduler.process(BatchProd(tasks, 2, lambda x: x)))
    test_scheduler.join()

    assert out == ["done-task1"]


def test_recurse(test_scheduler: Scheduler):

    tasks = [SchedulerTask(f"task{i}", requirements=[cap1]) for i in range(1, 9)]

    out = []
    for e in test_scheduler.process(ListProd(tasks, 10)):
        out.append(e)
        if len(out) % 2 == 0:
            tasks.append(tasks[len(out) // 2])

    test_scheduler.join()
    muls = [(1, 1), *[(i, 2) for i in range(2, 9)]]
    assert sorted(out) == sum([[f"done-task{i}"] * k for i, k in muls], [])
